/*
Copyright (C) 2009

This file is part of the NN applet
written by Max Bügler
http://www.maxbuegler.eu/

NN applet is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

NN applet is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
import javax.swing.*;
import java.util.ArrayList;
import java.awt.*;
import java.awt.event.MouseListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;

/**
 * Author: Max Buegler - max@sarstech.com
 * Date: Dec 4, 2009
 * Time: 9:00:27 PM
 */
public class DataPanel extends JPanel {
    private ArrayList<Point> points;
    private BufferedImage buffer;
    private double[][] field;
    private int color;

    public DataPanel(){
        points=new ArrayList<Point>();
        points.add(new Point(0,0,0));
        points.add(new Point(50,50,0));
        points.add(new Point(0,50,1));
        points.add(new Point(50,0,1));
        color=BufferedImage.TYPE_BYTE_INDEXED;
        Listener l=new Listener();
        addMouseListener(l);
        addMouseMotionListener(l);

    }

    public void setField(double[][] field) {
        this.field = field;
    }

    public ArrayList<Point> getPoints() {
        return points;
    }

    public void reset(){
        points.clear();
        field=null;
    }

    public void setRGB(boolean rgb){
        if (rgb)color=BufferedImage.TYPE_INT_RGB;
        else color=BufferedImage.TYPE_BYTE_INDEXED;
    }

    public void update(Graphics g){
        paint(g);
    }

    public void paintComponent(Graphics g){
        if (buffer==null||buffer.getWidth()!=getWidth()||buffer.getHeight()!=getHeight()||buffer.getType()!=color)
            buffer=new BufferedImage(getWidth(),getHeight(),color);
        Graphics2D g2=(Graphics2D)buffer.getGraphics();
        int cx=getWidth()/2,cy=getHeight()/2;
        if (field==null){
            g2.setColor(Color.WHITE);
            g2.fillRect(0,0,getWidth(),getHeight());
        }
        else{
            try{
                for (int x=0;x<field.length;x++){
                    for (int y=0;y<field[x].length;y++){
                        double pc=Math.max(0,Math.min(1,field[x][y]));
                        //buffer.setRGB(x,y,new Color((int)(127+(128*pc)),128,(int)(127+(128*(1-pc)))).getRGB());
                        g2.setColor(new Color((int)(127+(128*pc)),128,(int)(127+(128*(1-pc)))));
                        g2.fillRect(x,y,1,1);
                    }
                }
            }
            catch(Exception e){
                g2.setColor(Color.WHITE);
                g2.fillRect(0,0,getWidth(),getHeight());
            }
        }

        g2.setColor(Color.BLACK);
        g2.drawLine(0,cy,getWidth(),cy);
        g2.drawLine(cx,0,cx,getHeight());
        int x=1;
        while(cx+(x*50)<getWidth()){
            g2.drawLine(cx+(x*50),cy-5,cx+(x*50),cy+5);
            g2.drawLine(cx-(x*50),cy-5,cx-(x*50),cy+5);
            x++;
        }
        int y=1;
        while(cy+(y*50)<getHeight()){
            g2.drawLine(cx-5,cy+(y*50),cx+5,cy+(y*50));
            g2.drawLine(cx-5,cy-(y*50),cx+5,cy-(y*50));
            y++;
        }

        for (Point p:points){
            if (p.getC()!=-100){
                double pc=Math.max(0,Math.min(1,p.getC()));
                g2.setColor(new Color((int)(pc*255),0,(int)((1-pc)*255)));
                g2.fillOval(cx+p.getX()-9,cy+p.getY()-9,19,19);
            }
            if (p.getZ()>0.5)g2.setColor(Color.RED);
            else g2.setColor(Color.BLUE);
            g2.fillOval(cx+p.getX()-5,cy+p.getY()-5,11,11);
            g2.setColor(Color.BLACK);
            g2.drawOval(cx+p.getX()-5,cy+p.getY()-5,11,11);
        }
        g.drawImage(buffer,0,0,null);
    }
    class Listener implements MouseListener, MouseMotionListener {
        int dragindex=-1;
        public void mouseClicked(MouseEvent e) {
            int cx=getWidth()/2,cy=getHeight()/2;
            int z=1;
            if (e.getButton()==MouseEvent.BUTTON3)z=0;
            Point mp=new Point(e.getX()-cx,e.getY()-cy,z);
            for (int x=0;x<points.size();x++){
                if (points.get(x).distanceTo(mp)<5){
                    points.remove(x);
                    repaint();
                    return;
                }
            }
            points.add(mp);
            repaint();
        }

        public void mouseDragged(MouseEvent e) {
            if (dragindex<0)return;
            int cx=e.getX()-(getWidth()/2),cy=e.getY()-(getHeight()/2);
            points.get(dragindex).setX(cx);
            points.get(dragindex).setY(cy);
            repaint();
        }

        public void mouseMoved(MouseEvent e) {}

        public void mousePressed(MouseEvent e) {
            int cx=getWidth()/2,cy=getHeight()/2;
            if (e.getButton()!=MouseEvent.BUTTON1)return;
            Point mp=new Point(e.getX()-cx,e.getY()-cy);
            for (int x=0;x<points.size();x++){
                if (points.get(x).distanceTo(mp)<5){
                    dragindex=x;
                    return;
                }
            }
            dragindex=-1;
        }
        public void mouseReleased(MouseEvent e) {
            dragindex=-1;
        }
        public void mouseEntered(MouseEvent e) {}
        public void mouseExited(MouseEvent e) {}
    }
}
