/*
Copyright (C) 2009

This file is part of the NN applet
written by Max Bügler
http://www.maxbuegler.eu/

NN applet is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

NN applet is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
import NN.NeuralNetwork;
import NN.Matrix;

import javax.swing.*;
import java.applet.Applet;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.WindowListener;
import java.awt.event.WindowEvent;
import java.awt.event.ActionListener;
import java.util.Random;

/**
 * Author: Max Buegler - max@sarstech.com
 * Date: Dec 4, 2009
 * Time: 9:00:17 PM
 */
public class Main extends Applet {
    private JTextField alpha;
    private JTextField gamma;
    private JTextField epoch;
    private JTextField error;
    private TrainingThread thread;
    private JButton startstop;
    private TextFrame matrices;
    private NeuralNetwork nn;
    private DataPanel dp;
    private Visualizer vis;
    private Random rnd;
    private JCheckBox back;
    private JCheckBox backlr,backrgb,hardlim;
    private double[][] field;
    public void init(){
        dp=new DataPanel();
        vis=new Visualizer();
        rnd=new Random();
        setLayout(new BorderLayout());
        add(dp,BorderLayout.CENTER);
        matrices=new TextFrame("NN Matrices");
        final JTextField layers=new JTextField("2",5);
        nn=new NeuralNetwork(new int[]{2,2,1});
        alpha=new JTextField("0.2",5);
        gamma=new JTextField("0",5);
        epoch=new JTextField("0",5);
        error=new JTextField("unknown",15);
        epoch.setEditable(false);
        error.setEditable(false);
        JButton init=new JButton("Init");
        JButton show=new JButton("Show matrices");
        JButton visualize=new JButton("Visualize matrices");
        init.setFont(new Font("Arial",Font.PLAIN,10));
        show.setFont(new Font("Arial",Font.PLAIN,10));
        visualize.setFont(new Font("Arial",Font.PLAIN,10));
        JButton reset=new JButton("Reset");
        startstop=new JButton("Learn");
        JPanel toppanel=new JPanel();
        JPanel layerp=new JPanel();
        layerp.setLayout(new GridLayout(3,1));
        layerp.add(new JLabel("Layers:"));
        layerp.add(layers);
        hardlim=new JCheckBox("Hardlim output");
        layerp.add(hardlim);
        toppanel.add(layerp);
        JPanel initp=new JPanel();
        initp.setLayout(new GridLayout(3,1));
        initp.add(init);
        initp.add(show);
        initp.add(visualize);
        toppanel.add(initp);
        JPanel ststp=new JPanel();
        ststp.setLayout(new GridLayout(2,1));
        ststp.add(startstop);
        ststp.add(reset);
        toppanel.add(ststp);
        JPanel alphap=new JPanel();
        alphap.setLayout(new GridLayout(2,1));
        alphap.add(new JLabel("Alpha:"));
        alphap.add(alpha);
        toppanel.add(alphap);
        JPanel gammap=new JPanel();
        gammap.setLayout(new GridLayout(2,1));
        gammap.add(new JLabel("Gamma:"));
        gammap.add(gamma);
        toppanel.add(gammap);

        JPanel bottompanel=new JPanel();
        JPanel epochp=new JPanel();
        epochp.setLayout(new GridLayout(2,1));
        epochp.add(new JLabel("Epoch:"));
        epochp.add(epoch);
        bottompanel.add(epochp);
        JPanel errp=new JPanel();
        errp.setLayout(new GridLayout(2,1));
        errp.add(new JLabel("Error:"));
        errp.add(error);
        bottompanel.add(errp);
        JButton about=new JButton("About");
        JPanel backpanel=new JPanel();
        backpanel.setLayout(new GridLayout(3,1));
        back=new JCheckBox("Color back");
        backlr=new JCheckBox("Low resolution");
        backrgb=new JCheckBox("Full RGB");
        backpanel.add(back);
        backpanel.add(backlr);
        backpanel.add(backrgb);
        bottompanel.add(backpanel);
        bottompanel.add(about);
        add(toppanel,BorderLayout.NORTH);
        add(bottompanel,BorderLayout.SOUTH);
        thread=new TrainingThread();
        backrgb.addActionListener(new ActionListener(){
            public void actionPerformed(ActionEvent e) {
                dp.setRGB(backrgb.isSelected());
            }});

        back.addActionListener(new ActionListener(){
            public void actionPerformed(ActionEvent e) {
                if (!back.isSelected())
                    dp.setField(null);
            }});

        reset.addActionListener(new ActionListener(){
            public void actionPerformed(ActionEvent e) {
                dp.reset();
                dp.repaint();
            }});
        final Applet a=this;
        about.addActionListener(new ActionListener(){
            public void actionPerformed(ActionEvent e) {
                JOptionPane.showMessageDialog(a,"Written 2009 by Max Bügler. Visit http://www.sarstech.com");
            }});
        show.addActionListener(new ActionListener(){
            public void actionPerformed(ActionEvent e) {
                matrices.setVisible(true);
                if (thread.isStopped()){
                    matrices.setText(nn.toString());
                }
            }});
        visualize.addActionListener(new ActionListener(){
            public void actionPerformed(ActionEvent e) {
                vis.loadNN(nn);
                vis.setVisible(true);
                if (thread.isStopped()){
                    vis.updateNN();
                }
            }});

        init.addActionListener(new ActionListener(){
            public void actionPerformed(ActionEvent e) {
                thread.stopTraining();
                while(!thread.isStopped())try{Thread.sleep(5);}catch(Exception ex){}
                String s=layers.getText();
                int commas=0;
                for (int x=0;x<s.length();x++)
                    if (s.charAt(x)==',')commas++;

                int[] structure=new int[commas+3];
                structure[0]=2;
                int lastcomma=-1;
                int index=1;
                for (int x=0;x<s.length();x++){
                    if (s.charAt(x)==','){
                        structure[index]=Integer.parseInt(s.substring(lastcomma+1,x));
                        index++;
                        lastcomma=x;
                    }
                }
                structure[structure.length-2]=Integer.parseInt(s.substring(lastcomma+1));;
                structure[structure.length-1]=1;
                nn=new NeuralNetwork(structure);
                vis.loadNN(nn);

        }});
        startstop.addActionListener(new ActionListener(){
            public void actionPerformed(ActionEvent e) {
                if (thread.isStopped()){
                    thread=new TrainingThread();
                    thread.start();
                }else
                    thread.stopTraining();
            }});
    }
    class TrainingThread extends Thread{
        private boolean stop,stopped;
        public TrainingThread(){
            stopped=true;
            stop=true;
        }
        public void run(){
            stopped=false;
            stop=false;
            long ep=0;
            startstop.setText("Stop");
            while(!stop){
                try{
                    double a=Double.parseDouble(alpha.getText());
                    double g=Double.parseDouble(gamma.getText());
                    double err;
                    if (ep%10==0)err=trainRandomEpoch(a,g,true);
                    else err=trainRandomEpoch(a,g,false);
                    error.setText(err+"");
                    ep++;
                    epoch.setText(ep+"");
                    repaint();
                    if (matrices.isVisible())
                        matrices.setText(nn.toString());
                    if (vis.isVisible())
                        vis.updateNN();    
                    Thread.sleep(10);
                }
                catch(Exception e){}
            }
            startstop.setText("Learn");
            stopped=true;
        }
        public void stopTraining(){
            stop=true;
        }

        public boolean isStopped() {
            return stopped;
        }
    }
    private double trainRandomEpoch(double alpha, double gamma, boolean updateback){

        for (int x=0;x<dp.getPoints().size();x++){
            Point p=dp.getPoints().get(rnd.nextInt(dp.getPoints().size()));
            double[][] a0=new double[2][1];
            a0[0][0]=p.getX()/50.0;
            a0[1][0]=p.getY()/50.0;
            double[][] t=new double[1][1];
            t[0][0]=p.getZ();
            Matrix e=nn.backPropagate(new Matrix(a0),new Matrix(t),alpha,gamma,hardlim.isSelected());
        }
        double err=0;
        for (int x=0;x<dp.getPoints().size();x++){
            Point p=dp.getPoints().get(x);
            double[][] a0=new double[2][1];
            a0[0][0]=p.getX()/50.0;
            a0[1][0]=p.getY()/50.0;
            double res=nn.feedForward(new Matrix(a0),hardlim.isSelected()).getValue(0,0);
            p.setC(res);
            err+=Math.pow(p.getZ()-res,2);
        }
        if (back.isSelected()){
            if (updateback){
                field=new double[getWidth()][getHeight()];
                int cx=dp.getWidth()/2;
                int cy=dp.getHeight()/2;
                if (backlr.isSelected()){
                    for (int x=0;x<dp.getWidth()/4;x++){
                        for (int y=0;y<dp.getHeight()/4;y++){
                            double[][] a0=new double[2][1];
                            a0[0][0]=(4*x-cx+2)/50.0;
                            a0[1][0]=(4*y-cy+2)/50.0;
                            double val=nn.feedForward(new Matrix(a0),false).getValue(0,0);
                            for (int x1=0;x1<4;x1++)
                                for (int y1=0;y1<4;y1++)
                                    field[4*x+x1][4*y+y1]=val;
                        }
                    }
                }
                else{
                    for (int x=0;x<dp.getWidth();x++){
                        for (int y=0;y<dp.getHeight();y++){
                            double[][] a0=new double[2][1];
                            a0[0][0]=(x-cx)/50.0;
                            a0[1][0]=(y-cy)/50.0;
                            field[x][y]=nn.feedForward(new Matrix(a0),false).getValue(0,0);
                        }
                    }
                }
            }
        }
        else field=null;
        dp.setField(field);
        return err/dp.getPoints().size();
    }
    public void update(Graphics g){
        paint(g);
    }
}
