/*
Copyright (C) 2009

This file is part of the NN applet
written by Max Bügler
http://www.maxbuegler.eu/

NN applet is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

NN applet is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
package Space;



import java.awt.*;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

import javax.swing.JComponent;
import javax.swing.JFrame;

/**
 * Java2D 3D-Space
 * @author Max Buegler - Universiteit Maastricht - Knowledge Engineering 2007/2008 Block 1.3
 *
 */
public class Space2 extends JComponent{
	private ArrayList<Polygon> polygons;
    private ArrayList<SplineObject> splineObjects;
    private ArrayList<Label> labels;
	private Color backColor;
	private double[] camVector;
	private double[] layerXVector,layerYVector;
	private double[] camAngle;
	private double zoom;
	private boolean perspective=true;
	/**
	 * Constructs a new 3D Space
	 * @param back BackgroundColor
	 */
	public Space2(Color back){
        splineObjects= new ArrayList<SplineObject>();
		polygons= new ArrayList<Polygon>();
        labels=new ArrayList<Label>();
		backColor=back;
		layerYVector = new double[3];
		camVector = new double[3];
		layerXVector = new double[3];
		camAngle= new double[2];
		camAngle[0]=1;
		camAngle[1]=1;
		layerXVector[2]=0;
		zoom=1;
		calculateVectors();
		mouselist mL = new mouselist();
		addMouseListener(mL);
		addMouseMotionListener(mL);
		addMouseWheelListener(mL);
	}

    /**
	 * Clears Space
	 */
	public void clearSpace(){
		polygons= new ArrayList<Polygon>();
        splineObjects= new ArrayList<SplineObject>();
        labels=new ArrayList<Label>();
		//camAngle[0]=1;
		//camAngle[1]=1;
		//layerXVector[2]=0;
		//zoom=1;
		//calculateVectors();
	}
	public void setPerspektive(boolean p){
		perspective=p;
		repaint();
	}
	/**
	 * Adds a SplineObject to the 3D Space
	 * @param x SplineObject
	 */
	public void addSplineObject(SplineObject x){
		splineObjects.add(x);
	}

	public void addPolygon(Polygon x){
		polygons.add(x);
	}
    public void addLabel(Label x){
        labels.add(x);
    }

	/**
	 * Removes last SplineObject from the 3D Space
	 */
	public void remLastPolygon(){
		polygons.remove(polygons.size()-1);
	}
	
	public void paintComponent(Graphics g){
		Graphics2D g2=(Graphics2D)g;
		g2.drawImage(createView(), 0, 0, null);
	}
	
	public void update(Graphics g){
		paintComponent(g);
	}
	
	private void calculateVectors(){

		camVector[0]=Math.cos(camAngle[1])*Math.sin(camAngle[0]);
		camVector[1]=Math.cos(camAngle[1])*Math.cos(camAngle[0]);
		camVector[2]=Math.sin(camAngle[1]);
        layerXVector[0]=Math.cos(camAngle[0]);
        layerXVector[1]=-Math.sin(camAngle[0]);
        layerXVector[2]=0;
        layerYVector[0]=-Math.sin(camAngle[1])*Math.sin(camAngle[0]);
        layerYVector[1]=-Math.sin(camAngle[1])*Math.cos(camAngle[0]);
        layerYVector[2]=Math.cos(camAngle[1]);
 
	}
	
	private double[] getLayerIntersection(double[] n, int[] x){
		double nom=0,denom=0;
		double[] q ={(1000*n[0])-x[0],(1000*n[1])-x[1],(1000*n[2])-x[2]};
		if (perspective){
			nom=-x[0]*n[0]-x[1]*n[1]-x[2]*n[2];
			denom=q[0]*n[0]+q[1]*n[1]+q[2]*n[2];
		}
		else{
			nom=-(n[0]*x[0])+Math.pow(n[0],2)-(n[1]*x[1])+Math.pow(n[1],2)-(n[2]*x[2])+Math.pow(n[2],2);
			denom=(Math.pow(n[0],2)+Math.pow(n[1],2)+Math.pow(n[2],2));
		}
		double lamda=nom/denom;;
		double[] p = new double[3];
		if (perspective){
			p[0]=x[0]+lamda*q[0];
			p[1]=x[1]+lamda*q[1];
			p[2]=x[2]+lamda*q[2];
		}
		else{
			p[0]=x[0]+lamda*n[0];
			p[1]=x[1]+lamda*n[1];
			p[2]=x[2]+lamda*n[2];
		}
		return p;
	}
	
	private int[] getScreenPos(double[] n, int[] x){
		double[] p=getLayerIntersection(n, x);
                double[] a = layerXVector;
                double[] b = layerYVector;
                double[][] c1 = { {a[0], b[0]}, {a[1], b[1]}};//, {a[2], b[2],0} };
                double[][] c2 = { {a[0], b[0]}, {a[2], b[2]}};//, {a[2], b[2],0} };
                double[][] c3 = { {a[1], b[1]}, {a[2], b[2]}};//, {a[2], b[2],0} };
                double[][] d = { { p[0]}, { p[1]}};//, { p[2]} };
                Matrix e = new Matrix(c1);
                Matrix f = new Matrix(d);
                Matrix g;
                int[] pr = new int[2];
                pr[0]=0;
                pr[1]=0;
                try{
                    g = e.solve(f);
                    pr[0]=(int)(zoom*g.getValue(0, 0));
                    pr[1]=(int)(zoom*g.getValue(1, 0));
                    return pr;
                }
                catch(RuntimeException ex){}
                return pr;
	}
	
	private BufferedImage createView(){
        BufferedImage img = new BufferedImage(getWidth(),getHeight(),BufferedImage.TYPE_INT_ARGB);
		Graphics2D g2 = img.createGraphics();
		g2.setColor(backColor);
        Collections.sort(polygons,new Comparator<Polygon>(){
            public int compare(Polygon o1, Polygon o2) {
                double[] avg1=o1.getAverage();
                double d1=Math.sqrt(Math.pow(1000*camVector[0]-avg1[0],2)+Math.pow(1000*camVector[1]-avg1[1],2)+Math.pow(1000*camVector[2]-avg1[2],2));
                double[] avg2=o2.getAverage();
                double d2=Math.sqrt(Math.pow(1000*camVector[0]-avg2[0],2)+Math.pow(1000*camVector[1]-avg2[1],2)+Math.pow(100*camVector[2]-avg2[2],2));
                return (int)Math.round(100*(d1-d2));
            }
        });
		g2.fillRect(0, 0, getWidth(), getHeight());
		int[] center = {img.getWidth()/2,img.getHeight()/2};
		for (int x=0;x<polygons.size();x++){
			try{
				Polygon current = polygons.get(x);
                int[] xPoints=new int[current.getX().length],yPoints=new int[current.getX().length];
				for (int y=0;y<current.getX().length;y++){
			        int[] p=getScreenPos(camVector,new int[]{current.getX()[y],current.getY()[y],current.getZ()[y]});
			        p[0]+=center[0];
			        p[1]+=center[1];
                    xPoints[y]=p[0];
                    yPoints[y]=p[1];
				}
                g2.setColor(current.getCol());
                //if (x==0)g2.setStroke(new BasicStroke(2));
                //else g2.setStroke(new BasicStroke(1));
                //g2.drawLine(start[0], start[1],end[0], end[1]);
                g2.fillPolygon(xPoints,yPoints,xPoints.length);
			}
			catch (Exception e){}
		}
		for (int x=splineObjects.size()-1;x>=0;x--){
			try{
				SplineObject current = splineObjects.get(x);
				for (int y=0;y<current.getSplines().size();y++){
					Spline currentS = current.getSplines().get(y);
					int[] start;
					int[] end;
			        start=getScreenPos(camVector,currentS.getP1());
			        start[0]+=center[0];
			        start[1]+=center[1];
			        end=getScreenPos(camVector,currentS.getP2());
			        end[0]+=center[0];
			        end[1]+=center[1];
					g2.setColor(current.getSplines().get(y).getC1());
					if (x==0)g2.setStroke(new BasicStroke(2));
					else g2.setStroke(new BasicStroke(1));
					g2.drawLine(start[0], start[1],end[0], end[1]);
				}
			}
			catch (Exception e){}
		}
        for (int x=0;x<labels.size();x++){
            try{
                int[] loc=getScreenPos(camVector,labels.get(x).getP());
                loc[0]+=center[0];
                loc[1]+=center[0];
                g2.setColor(labels.get(x).getCol());
                g2.drawString(labels.get(x).getText(),loc[0],loc[1]);
            }
            catch(Exception e){}
        }
		return img;
	}
	
	/**
	 * Universal MouseListener (one for all)
	 * @author Max Buegler - Universiteit Maastricht - Knowledge Engineering 2007/2008 Block 1.3
	 *
	 */
	class mouselist extends MouseAdapter implements MouseMotionListener,MouseWheelListener{
		int clickedX, clickedY;
		public void mousePressed(MouseEvent e){
			clickedX=e.getX();
			clickedY=e.getY();
		}
		public void mouseMoved(MouseEvent e){}
		public void mouseDragged(MouseEvent e){
			int delta=e.getX()-clickedX;
			camAngle[0]-=delta*0.01;
			delta=e.getY()-clickedY;
			camAngle[1]-=delta*0.01;
			clickedX=e.getX();
			clickedY=e.getY();
			calculateVectors();
			repaint();			
		}
		public void mouseWheelMoved(MouseWheelEvent e){
			zoom-=e.getWheelRotation()*0.5;
			repaint();
		}
	}


    public static void main(String[] args){
        JFrame frame=new JFrame("FUN");
        frame.setSize(400,400);
        Space2 s=new Space2(Color.WHITE);
        s.addPolygon(new Polygon(new int[]{0,10,20},new int[]{0,10,0},new int[]{0,0,0},Color.RED));
        s.addPolygon(new Polygon(new int[]{0,10,20},new int[]{0,10,0},new int[]{20,20,20},Color.GREEN));
        s.addPolygon(new Polygon(new int[]{0,10,20},new int[]{0,10,0},new int[]{10,10,10},Color.BLUE));
        frame.setLayout(new GridLayout(1,1));
        frame.add(s);
        frame.setVisible(true);
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    }
}
