/*
Copyright (C) 2009

This file is part of the NN applet
written by Max Bügler
http://www.maxbuegler.eu/

NN applet is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

NN applet is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
import NN.NeuralNetwork;

import javax.swing.*;
import java.awt.*;
import java.util.ArrayList;

import Space.*;
import Space.Polygon;
import Space.Label;
import NN.Matrix;

/**
 * Author: Max Buegler - max@sarstech.com
 * Date: Dec 11, 2009
 * Time: 11:05:52 AM
 */
public class Visualizer extends JFrame {
    private NeuralNetwork nn;
    private Space2[] bias,weights;
    private boolean solid;
    private JPanel viewport;

    public Visualizer(){
        super("NN Visualizer");
        setSize(500,400);
        solid=false;
        viewport=new JPanel();
        super.setLayout(new GridLayout(1,1));
        super.add(viewport);
    }

    public boolean isSolid() {
        return solid;
    }

    public void setSolid(boolean solid) {
        this.solid = solid;
    }

    public void loadNN(NeuralNetwork nn){
        remove(viewport);
        viewport=new JPanel();
        viewport.setLayout(new GridLayout(2,nn.getWeights().length));
        weights=new Space2[nn.getWeights().length];
        bias=new Space2[nn.getBias().length];
        this.nn=nn;
        for (int x=0;x<nn.getWeights().length;x++){
            Matrix w=nn.getWeights()[x];
            Matrix b=nn.getBias()[x];
            weights[x]=new Space2(Color.WHITE);
            bias[x]=new Space2(Color.WHITE);
            if (solid){
                ArrayList<Polygon> polys=matrixToPolygonObject(w);
                for (Polygon p:polys)weights[x].addPolygon(p);
                polys=matrixToPolygonObject(b);
                for (Polygon p:polys)bias[x].addPolygon(p);
            }
            else{
                weights[x].addSplineObject(matrixToSplineObject(w));
                bias[x].addSplineObject(matrixToSplineObject(b));
            }
        }
        for (int x=0;x<nn.getWeights().length;x++)
            viewport.add(weights[x]);
        for (int x=0;x<nn.getWeights().length;x++)
            viewport.add(bias[x]);
        add(viewport);
    }
    public void updateNN(){
        for (int x=0;x<nn.getWeights().length;x++){
            Matrix w=nn.getWeights()[x];
            Matrix b=nn.getBias()[x];
            weights[x].clearSpace();
            bias[x].clearSpace();
            if (solid){
                ArrayList<Polygon> polys=matrixToPolygonObject(w);
                for (Polygon p:polys)weights[x].addPolygon(p);
            }
            else
                weights[x].addSplineObject(matrixToSplineObject(w));
            bias[x].addSplineObject(matrixToSplineObject(b));
            weights[x].addLabel(new Label(new int[]{-10*w.getW()/2,-10*w.getH()/2,0},"(0,0)",Color.BLACK));
            bias[x].addLabel(new Label(new int[]{-10*b.getW()/2,-10*b.getH()/2,0},"(0,0)",Color.BLACK));
            weights[x].addLabel(new Label(new int[]{-10*w.getW()/2,10*w.getH()/2,0},"(1,0)",Color.BLACK));
            bias[x].addLabel(new Label(new int[]{-10*b.getW()/2,10*b.getH()/2,0},"(1,0)",Color.BLACK));
        }
        repaint();
    }
    private static SplineObject matrixToSplineObject(Matrix m){
        SplineObject out=new SplineObject();
        int cx=m.getW()/2;
        int cy=m.getH()/2;
        for (int x=0;x<m.getW();x++){
            for (int y=0;y<m.getH();y++){
                if (x<m.getW()-1)out.addSpline(new Spline(new int[]{10*(x-cx),10*(y-cy),(int)Math.round(10*m.getValue(y,x))},new int[]{10*(x+1-cx),10*(y-cy),(int)Math.round(10*m.getValue(y,x+1))},getColor(m.getValue(y,x+1))));
                if (y<m.getH()-1)out.addSpline(new Spline(new int[]{10*(x-cx),10*(y-cy),(int)Math.round(10*m.getValue(y,x))},new int[]{10*(x-cx),10*(y+1-cy),(int)Math.round(10*m.getValue(y+1,x))},getColor(m.getValue(y+1,x))));
            }
        }
        return out;
    }
    private static ArrayList<Polygon> matrixToPolygonObject(Matrix m){
        ArrayList<Polygon> out=new ArrayList<Polygon>();
        int cx=m.getW()/2;
        int cy=m.getH()/2;
        for (int x=0;x<m.getW()-1;x++){
            for (int y=0;y<m.getH()-1;y++){
                out.add(new Polygon(new int[]{10*(x-cx),10*(x+1-cx),10*(x+1-cx),10*(x-cx)},new int[]{10*(y-cy),10*(y-cy),10*(y+1-cy),10*(y+1-cy)},new int[]{(int)Math.round(10*m.getValue(y,x)),(int)Math.round(10*m.getValue(y,x+1)),(int)Math.round(10*m.getValue(y+1,x+1)),(int)Math.round(10*m.getValue(y+1,x))},getColor(m.getValue(y,x))));
            }
        }
        return out;
    }
    private static Color getColor(double v){
        int g=(int)Math.round(255*Math.exp(-Math.pow(v,2)/2.0));
        int r=0;
        if (v>0)r=255-g;
        int b=0;
        if (v<0)b=255-g;
        return new Color(r,g,b);
    }
}
